import copy
import logging
import math
from typing import List, Dict, Any
from uuid import UUID

from argus.backend.db import ScyllaCluster
from argus.backend.models.result import ArgusGenericResultMetadata, ArgusGenericResultData

LOGGER = logging.getLogger(__name__)

default_options = {
    "scales": {
        "y": {
            "beginAtZero": True,
            "title": {
                "display": True,
                "text": ''
            }
        },
        "x": {
            "type": "time",
            "time": {
                "unit": "day",
                "displayFormats": {
                    "day": "yyyy-MM-dd",
                },
            },
            "title": {
                "display": True,
                "text": 'SUT Date'
            }
        },
    },
    "elements": {
        "line": {
            "tension": .1,
        }
    },
    "plugins": {
        "legend": {
            "position": 'top',
        },
        "title": {
            "display": True,
            "text": ''
        }
    }
}

colors = [
    'rgba(255, 0, 0, 1.0)',  # Red
    'rgba(0, 255, 0, 1.0)',  # Green
    'rgba(0, 0, 255, 1.0)',  # Blue
    'rgba(0, 255, 255, 1.0)',  # Cyan
    'rgba(255, 0, 255, 1.0)',  # Magenta
    'rgba(255, 255, 0, 1.0)',  # Yellow
    'rgba(255, 165, 0, 1.0)',  # Orange
    'rgba(128, 0, 128, 1.0)',  # Purple
    'rgba(50, 205, 50, 1.0)',  # Lime
    'rgba(255, 192, 203, 1.0)',  # Pink
    'rgba(0, 128, 128, 1.0)',  # Teal
    'rgba(165, 42, 42, 1.0)',  # Brown
    'rgba(0, 0, 128, 1.0)',  # Navy
    'rgba(128, 128, 0, 1.0)',  # Olive
    'rgba(255, 127, 80, 1.0)'  # Coral
]


def get_sorted_data_for_column_and_row(data: List[Dict[str, Any]], column: str, row: str) -> List[Dict[str, Any]]:
    return sorted([{"x": entry.sut_timestamp.strftime('%Y-%m-%dT%H:%M:%SZ'),
                    "y": entry.value,
                    "id": entry.run_id}
                   for entry in data if entry['column'] == column and entry['row'] == row],
                  key=lambda x: x['x'])


def get_min_max_y(datasets: List[Dict[str, Any]]) -> (float, float):
    """0.5 - 1.5 of min/max of 50% results"""
    y = [entry['y'] for dataset in datasets for entry in dataset['data']]
    if not y:
        return 0, 0
    sorted_y = sorted(y)
    lower_percentile_index = int(0.25 * len(sorted_y))
    upper_percentile_index = int(0.75 * len(sorted_y)) - 1
    y_min = sorted_y[lower_percentile_index]
    y_max = sorted_y[upper_percentile_index]
    return math.floor(0.5 * y_min), math.ceil(1.5 * y_max)


def round_datasets_to_min_max(datasets: List[Dict[str, Any]], min_y: float, max_y: float) -> List[Dict[str, Any]]:
    """Round values to min/max and provide original value for tooltip"""
    for dataset in datasets:
        for entry in dataset['data']:
            val = entry['y']
            if val > max_y:
                entry['y'] = max_y
                entry['ori'] = val
            elif val < min_y:
                entry['y'] = min_y
                entry['ori'] = val
    return datasets


def create_chartjs(table, data):
    graphs = []
    for column in table.columns_meta:
        datasets = [
            {"label": row,
             "borderColor": colors[idx % len(colors)],
             "borderWidth": 3,
             "showLine": True,
             "data": get_sorted_data_for_column_and_row(data, column.name, row)} for idx, row in enumerate(table.rows_meta)]
        min_y, max_y = get_min_max_y(datasets)
        datasets = round_datasets_to_min_max(datasets, min_y, max_y)
        if not min_y + max_y:
            # filter out those without data
            continue
        options = copy.deepcopy(default_options)
        options["plugins"]["title"]["text"] = f"{table.name} - {column.name}"
        options["scales"]["y"]["title"]["text"] = f"[{column.unit}]"
        options["scales"]["y"]["min"] = min_y
        options["scales"]["y"]["max"] = max_y
        graphs.append({"options": options, "data":
            {"datasets": datasets}})
    return graphs


class ResultsService:

    def __init__(self, database_session=None):
        self.session = database_session if database_session else ScyllaCluster.get_session()

    def get_results(self, test_id: UUID):
        graphs = []
        res = ArgusGenericResultMetadata.objects(test_id=test_id).all()
        for table in res:
            data = ArgusGenericResultData.objects(test_id=test_id, name=table.name).all()
            graphs.extend(create_chartjs(table, data))
        return graphs



from dataclasses import dataclass
from typing import TypedDict
from uuid import UUID


class RawMatrixTestCase(TypedDict):
    name: str
    status: str
    time: float
    classname: str
    message: str


class RawMatrixTestSuite(TypedDict):
    name: str
    tests: int
    failures: int
    disabled: int
    skipped: int
    passed: int
    errors: int
    time: float
    cases: list[RawMatrixTestCase]


class RawMatrixTestResult(TypedDict):
    name: str
    driver_name: str
    tests: int
    failures: int
    errors: int
    disabled: int
    skipped: int
    passed: int
    time: float
    timestamp: str
    suites: list[RawMatrixTestSuite]


@dataclass(init=True, frozen=True)
class DriverMatrixSubmitResultRequest():
    schema_version: str
    run_id: UUID
    driver_type: str
    driver_name: str
    raw_xml: str


@dataclass(init=True, frozen=True)
class DriverMatrixSubmitFailureRequest():
    schema_version: str
    run_id: UUID
    driver_type: str
    driver_name: str
    failure_reason: str


@dataclass(init=True, frozen=True)
class DriverMatrixSubmitEnvRequest():
    schema_version: str
    run_id: UUID
    raw_env: str



from flask import Blueprint, request

from argus.backend.error_handlers import handle_api_exception
from argus.backend.plugins.driver_matrix_tests.raw_types import DriverMatrixSubmitEnvRequest, DriverMatrixSubmitFailureRequest, DriverMatrixSubmitResultRequest
from argus.backend.service.user import api_login_required
from argus.backend.plugins.driver_matrix_tests.service import DriverMatrixService
from argus.backend.util.common import get_payload

bp = Blueprint("driver_matrix_api", __name__, url_prefix="/driver_matrix")
bp.register_error_handler(Exception, handle_api_exception)


@bp.route("/test_report", methods=["GET"])
@api_login_required
def driver_matrix_test_report():

    build_id = request.args.get("buildId")
    if not build_id:
        raise Exception("No build id provided")
    

    result = DriverMatrixService().tested_versions_report(build_id=build_id)
    return {
        "status": "ok",
        "response": result
    }

@bp.route("/result/submit", methods=["POST"])
@api_login_required
def submit_result():
    payload = get_payload(request)
    request_data = DriverMatrixSubmitResultRequest(**payload)

    result = DriverMatrixService().submit_driver_result(driver_name=request_data.driver_name, driver_type=request_data.driver_type, run_id=request_data.run_id, raw_xml=request_data.raw_xml)
    return {
        "status": "ok",
        "response": result
    }


@bp.route("/result/fail", methods=["POST"])
@api_login_required
def submit_failure():
    payload = get_payload(request)
    request_data = DriverMatrixSubmitFailureRequest(**payload)

    result = DriverMatrixService().submit_driver_failure(driver_name=request_data.driver_name, driver_type=request_data.driver_type, run_id=request_data.run_id, failure_reason=request_data.failure_reason)
    return {
        "status": "ok",
        "response": result
    }

@bp.route("/env/submit", methods=["POST"])
@api_login_required
def submit_env():
    payload = get_payload(request)
    request_data = DriverMatrixSubmitEnvRequest(**payload)

    result = DriverMatrixService().submit_env_info(run_id=request_data.run_id, raw_env=request_data.raw_env)
    return {
        "status": "ok",
        "response": result
    }
